# https://bbpteam.epfl.ch/project/spaces/display/BGLIB/Neurodamus
# Copyright 2005-2021 Blue Brain Project, EPFL. All rights reserved.
"""Implements applying modifications that mimic experimental manipulations

New Modification classes must be registered, using the appropriate decorator.
Also, when instantiated by the framework, __init__ is passed three arguments
(1) target (2) mod_info: dict (3) cell_manager. Example

>>> @ModificationManager.register_type
>>> class TTX:
>>>
>>> def __init__(self, target, mod_info: dict, cell_manager):
>>>     tpoints = target.get_point_list(cell_manager, section_type, compartment_type)
>>>     for point in tpoints:
>>>         for sec_id, sc in enumerate(point.sclst):
>>>             if not sc.exists():
>>>                 continue
>>>             sec = sc.sec

"""

import ast
import logging

from .core import NeuronWrapper as Nd
from .core.configuration import ConfigurationError
from .utils.logging import log_verbose
from neurodamus.report_parameters import CompartmentType, SectionType


class ModificationManager:
    """A manager for circuit Modifications.
    Overrides HOC manager, as the only Modification there (TTX) is outdated.
    """

    _mod_types = {}  # modification handled in Python

    def __init__(self, target_manager):
        self._target_manager = target_manager
        self._modifications = []

    def interpret(self, target_spec, mod_info):
        mod_t_name = mod_info["Type"]
        mod_t = self._mod_types.get(mod_t_name)
        if not mod_t:
            raise ConfigurationError(f"Unknown Modification {mod_t_name}")
        target = self._target_manager.get_target(target_spec)
        cell_manager = self._target_manager._cell_manager
        mod = mod_t(target, mod_info, cell_manager)
        self._modifications.append(mod)

    @classmethod
    def register_type(cls, mod_class):
        """Registers a new class as a handler for a new modification type"""
        cls._mod_types[mod_class.__name__] = mod_class
        return mod_class


@ModificationManager.register_type
class TTX:
    """Applies sodium channel block to all sections of the cells in the given target

    Uses TTXDynamicsSwitch as in BGLibPy. Overrides HOC version, which is outdated
    """

    def __init__(self, target, mod_info: dict, cell_manager):
        tpoints = target.get_point_list(
            cell_manager, section_type=SectionType.ALL, compartment_type=CompartmentType.ALL
        )

        # insert and activate TTX mechanism in all sections of each cell in target
        for tpoint_list in tpoints:
            for sc in tpoint_list.sclst:
                if not sc.exists():  # skip sections not on this split
                    continue
                sec = sc.sec
                if not Nd.ismembrane("TTXDynamicsSwitch", sec=sec):
                    sec.insert("TTXDynamicsSwitch")
                sec.ttxo_level_TTXDynamicsSwitch = 1.0


@ModificationManager.register_type
class ConfigureAllSections:
    """Perform one or more assignments involving section attributes,
    for all sections that have all the referenced attributes.

    Use case is modifying mechanism variables from config.
    """

    def __init__(self, target, mod_info: dict, cell_manager):
        config, config_attrs = self.parse_section_config(mod_info["SectionConfigure"])
        tpoints = target.get_point_list(
            cell_manager, section_type=SectionType.ALL, compartment_type=CompartmentType.ALL
        )

        napply = 0  # number of sections where config applies
        # change mechanism variable in all sections that have it
        for tpoint_list in tpoints:
            for _, sc in enumerate(tpoint_list.sclst):
                if not sc.exists():  # skip sections not on this split
                    continue
                sec = sc.sec
                if all(hasattr(sec, x) for x in config_attrs):  # if has all attributes
                    # unsafe but sanitized
                    exec(config, {"__builtins__": None}, {"sec": sec})  # noqa: S102
                    napply += 1

        log_verbose(f"Applied to {napply} sections")

        if napply == 0:
            logging.warning(
                "ConfigureAllSections applied to zero sections, "
                "please check its SectionConfigure for possible mistakes"
            )

    def parse_section_config(self, config):
        config = config.replace("%s.", "__sec_wildcard__.")  # wildcard to placeholder
        all_attrs = self.AttributeCollector()
        tree = ast.parse(config)
        for elem in tree.body:  # for each semicolon-separated statement
            # check assignment targets
            for tgt in self.assignment_targets(elem):
                # must be single assignment of a __sec_wildcard__ attribute
                if not isinstance(tgt, ast.Attribute) or tgt.value.id != "__sec_wildcard__":
                    raise ConfigurationError(
                        "SectionConfigure only supports single assignments "
                        "of attributes of the section wildcard %s"
                    )
            all_attrs.visit(elem)  # collect attributes in assignment
        config = config.replace("__sec_wildcard__.", "sec.")  # placeholder to section variable

        return config, all_attrs.attrs

    class AttributeCollector(ast.NodeVisitor):
        """Node visitor collecting all attribute names in a set"""

        attrs = set()

        def visit_Attribute(self, node):
            self.attrs.add(node.attr)

    @staticmethod
    def assignment_targets(node):
        if isinstance(node, ast.Assign):
            return node.targets
        if isinstance(node, ast.AugAssign):
            return [node.target]
        raise ConfigurationError(
            "SectionConfigure must consist of one or more semicolon-separated assignments"
        )



"""Neurodamus is a software for handling neuronal simulation using neuron.

Copyright (c) 2018 Blue Brain Project, EPFL.
All rights reserved
"""

import logging
import sys

from neuron import h

from neurodamus import commands
from neurodamus.utils.cli import extract_arguments


def main():
    args = []
    try:
        args = extract_arguments(sys.argv)
    except ValueError:
        logging.exception()
        return 1

    return commands.neurodamus(args)


if __name__ == "__main__":
    # Returns exit code and calls MPI.Finalize
    h.quit(main())



"""Module implementing entry functions"""

import logging
import os
import sys
import time
from pathlib import Path

from docopt import docopt

from .core import MPI, OtherRankError
from .core.configuration import EXCEPTION_NODE_FILENAME, ConfigurationError, LogLevel
from .utils.pyutils import docopt_sanitize
from neurodamus.node import Neurodamus
from neurodamus.utils.timeit import TimerManager


def neurodamus(args=None):
    """Neurodamus

    Usage:
        neurodamus <ConfigFile> [options]
        neurodamus --version
        neurodamus --help

    Options:
        -v --verbose            Increase verbosity level
        --debug                 Extremely verbose mode for debugging
        --disable-reports       Disable all reports [default: False]
        --build-model=[AUTO, ON, OFF]
                                Shall it build and eventually overwrite? [default: AUTO]
                                - AUTO: build model if doesn't exist and simulator is coreneuron
                                - ON: always build and eventually overwrite the model
                                - OFF: Don't build the model. Simulation may fail to start
        --simulate-model=[ON, OFF]
                                Shall the simulation start automatically? [default: ON]
        --output-path=PATH      Alternative output directory, overriding the config file's
        --keep-build            Keep coreneuron intermediate data in a folder named `build`.
                                Otherwise deleted at the end. ``--save=<PATH>`` overrides this.
        --modelbuilding-steps=<number>
                                Set the number of ModelBuildingSteps for the CoreNeuron sim
        --experimental-stims    Shall use only Python stimuli? [default: False]
        --lb-mode=[RoundRobin, WholeCell, MultiSplit, Memory]
                                The Load Balance mode.
                                - RoundRobin: Disable load balancing. Good for quick simulations
                                - WholeCell: Does a first pass to compute load balancing and
                                    redistributes cells so that CPU load is similar among ranks
                                - MultiSplit: Allows splitting cells into pieces for distribution.
                                    WARNING: This mode is incompatible with CoreNeuron
                                - Memory: Load balance based on memory usage. By default, it uses
                                    the "allocation_r#_c#.pkl.gz" file to load a pre-computed load
                                    balance
        --save=<PATH>           Path to create a save point (at tstop) to enable restore. Only
                                available for CoreNEURON.
        --restore=<PATH>        Restore and resume simulation from a save point. Only available
                                for CoreNEURON.
        --dump-cell-state=<GID(s)>
                                Dump cell state debug files on tstart and tstop.
                                For NEURON, accepts a list of GIDs or ranges (e.g., 1,2,3-6,9).
                                For CoreNEURON, behavior is unchanged and only one GID is accepted.
                                If a list is provided, only the first GID will be used.
        --enable-shm=[ON, OFF]  Enables the use of /dev/shm for coreneuron_input (available
                                only on linux) [default: OFF]
        --model-stats           Show model stats in CoreNEURON simulations [default: False]
        --dry-run               Dry-run simulation to estimate memory usage [default: False]
        --crash-test            Run the simulation with single section cells and single synapses
        --num-target-ranks=<number>  Number of ranks to target for dry-run load balancing
        --coreneuron-direct-mode     Run CoreNeuron in direct memory mode transfered from Neuron,
                                     without writing model data to disk.
        --use-color=[ON, OFF]  If OFF, forces no color to be used in logs; [default: ON]
        --report-buffer-size=<number> Override the size in MB each rank will allocate for each
                                      report buffer to hold data. When the buffer is full, the
                                      ranks will aggregate data for writing to disk. Default: 8 MB
        --cell-permute=[unpermuted, node-adjacency]   Cell permutation [default: unpermuted].
                                Only available for CoreNEURON.
                                Currently incompatible with NEURON. Options:
                                - unpermuted: No permutation
                                - node-adjacency: Optimise for node adjacency
    """
    from . import __version__

    options = docopt_sanitize(docopt(neurodamus.__doc__, args, version=__version__))
    config_file = options.pop("ConfigFile")
    log_level = _pop_log_level(options)

    if not os.path.isfile(config_file):
        logging.error("Config file not found: %s", config_file)
        return 1

    # Shall replace process with special? Don't if is special or already replaced
    if not sys.argv[0].endswith("special") and not os.environ.get("NEURODAMUS_SPECIAL"):
        _attempt_launch_special(config_file)

    # Warning control before starting the process
    _filter_warnings()

    # Some previous executions may have left a bad exception node file
    # This is done now so it's a very early stage and we know the mpi rank
    if MPI.rank == 0 and os.path.exists(EXCEPTION_NODE_FILENAME):
        os.remove(EXCEPTION_NODE_FILENAME)

    try:
        Neurodamus(config_file, auto_init=True, logging_level=log_level, **options).run()
        TimerManager.timeit_show_stats()
    except ConfigurationError:  # Common, only show error in Rank 0
        if MPI._rank == 0:  # Use _rank so that we avoid init
            logging.exception("ConfigurationError")
        return 1
    except OtherRankError:
        return 1  # no need for _mpi_abort, error is being handled by all ranks
    except:  # noqa: E722
        show_exception_abort("Unhandled Exception. Terminating...", sys.exc_info())
        return 1
    return 0


def _pop_log_level(options):
    log_level = LogLevel.DEFAULT
    if options.pop("debug", False):
        log_level = LogLevel.DEBUG
    elif options.pop("verbose", False):
        log_level = LogLevel.VERBOSE

    if log_level >= LogLevel.VERBOSE:
        from pprint import pprint

        pprint(options)  # noqa: T203

    return log_level


def show_exception_abort(err_msg, exc_info):
    """Show an exception info in only one rank

    Several ranks are likely to be in sync so a simple touch wont work.
    Processes that dont see any file will register (append) their rank id
    First one is elected to print
    """
    err_file = Path(EXCEPTION_NODE_FILENAME)
    ALL_RANKS_SYNC_WINDOW = 1

    if err_file.exists():
        return

    with open(err_file, "a", encoding="utf-8") as f:
        f.write(str(MPI.rank) + "\n")

    with open(err_file, encoding="utf-8") as f:
        line0 = f.readline().strip()

    if str(MPI.rank) == line0:
        logging.critical(err_msg, exc_info=exc_info)

    time.sleep(ALL_RANKS_SYNC_WINDOW)  # give time to the rank that logs the exception
    _mpi_abort()  # abort all ranks which have waited. Seems to help avoiding MPT stack


def _attempt_launch_special(config_file):
    import shutil

    special = shutil.which("special")
    if os.path.isfile("x86_64/special"):  # prefer locally compiled special
        special = os.path.abspath("x86_64/special")
    if special is None:
        logging.warning(
            "special not found. Running neurodamus from Python with libnrnmech. "
            "-> DO NOT USE WITH PRODUCTION RUNS"
        )
        return
    neurodamus_py_root = os.environ.get("NEURODAMUS_PYTHON")
    if not neurodamus_py_root:
        logging.warning(
            "No NEURODAMUS_PYTHON set. Running neurodamus from Python with libnrnmech. "
            "-> DO NOT USE WITH PRODUCTION RUNS"
        )
        return
    print("::INIT:: Special available. Replacing binary...")  # noqa: T201
    os.environ["NEURODAMUS_SPECIAL"] = "1"
    init_script = os.path.join(neurodamus_py_root, "init.py")
    os.execl(special, "-mpi", "-python", init_script, "--configFile=" + config_file, *sys.argv[2:])


def _mpi_abort():
    import ctypes

    c_api = ctypes.CDLL(None)
    c_api.MPI_Abort(0)


def _filter_warnings():
    """Control matched warning to display once in rank 0.

    Warning 1:
    "special" binaries built with %intel build_type=Release,RelWithDebInfo flushes
    denormal results to zero, which triggers the numpy warning for subnormal in every rank.
    Reduce this type of warning displayed once in rank0.
    Note: "special" with build_type = FastDebug/Debug or calling the simulation process
       in python (built with gcc) does not have such flush-to-zero warning.
    """
    import warnings

    action = "once" if MPI.rank == 0 else "ignore"

    warnings.filterwarnings(
        action=action,
        message="The value of the smallest subnormal for .* type is zero.",
        category=UserWarning,
        module="numpy",
    )



import itertools
import logging
from collections import defaultdict
from collections.abc import Iterator
from functools import lru_cache, reduce
from operator import or_

import libsonata
import numpy as np

from .core import NeuronWrapper as Nd
from .core.configuration import ConfigurationError
from .core.nodeset import SelectionNodeSet
from .report_parameters import CompartmentType, SectionType
from .utils import compat
from .utils.logging import log_verbose


class TargetError(Exception):
    """A Exception class specific to data error with targets and nodesets"""


class TargetPointList:
    """List of target points (TPoint) in a neuron.

    Each TPoint is defined by a triplet: (gid, sec_ref, x) where:
      gid: neuron identifier
      sec_ref: reference to the section
      x (in [0, 1]): offset along the section

    The object keeps also track of the respective section_ids for convenience since
    there is no easy way to find the section id from its reference.

    Maintains parallel lists of section IDs, section references, and offsets,
    with methods to append, extend, validate, and iterate over the stored points.
    """

    def __init__(self, gid: int):
        self.gid: int = gid
        self.sclst_ids: list = []  # List of section ids
        self.sclst: list = []  # List of section references
        self.x: list = []  # List of point values

    def append(self, section_id: int, section: object, point: float) -> None:
        self.x.append(point)
        self.sclst.append(section)
        self.sclst_ids.append(section_id)

    def extend(self, other: "TargetPointList") -> None:
        assert isinstance(other, TargetPointList), (
            f"Expected TargetPointList, got {type(other).__name__}"
        )
        self.x.extend(other.x)
        self.sclst.extend(other.sclst)
        self.sclst_ids.extend(other.sclst_ids)

    def validate(self) -> None:
        if len(self.x) != len(self.sclst) != len(self.sclst_ids):
            raise RuntimeError(
                f"TargetPointList invariant violated: "
                f"x has {len(self.x)} elements, "
                f"sclst has {len(self.sclst)} elements, "
                f"sclst_ids has {len(self.sclst_ids)} elements. "
                f"Expected all lists to have equal length."
            )

    def __len__(self) -> int:
        self.validate()
        return len(self.sclst)

    def __iter__(self) -> Iterator[tuple[int, object, float]]:
        self.validate()
        return iter(zip(self.sclst_ids, self.sclst, self.x))

    def __str__(self) -> str:
        self.validate()
        ids = np.array(self.sclst_ids)
        xs = np.array(self.x)
        return f"TargetPointList(gid={self.gid}, size={len(self)}):\n  ids: {ids}\n  x:   {xs}"


class TargetSpec:
    """Definition of a new-style target, accounting for multipopulation"""

    GLOBAL_TARGET_NAME = "_ALL_"

    def __init__(self, target_name):
        """Initialize a target specification

        Args:
            target_name: the target name. For specifying a population use
                the format ``population:target_name``
        """
        if target_name and ":" in target_name:
            self.population, self.name = target_name.split(":")
        else:
            self.name = target_name
            self.population = None
        if not self.name:
            self.name = None

    def __str__(self):
        return (
            (self.name or "")
            if self.population is None
            else "{}:{}".format(self.population, self.name or "")
        )

    def __repr__(self):
        return "<TargetSpec: " + str(self) + ">"

    @property
    def simple_name(self):
        if self.name is None and self.population is None:
            return self.GLOBAL_TARGET_NAME
        return str(self).replace(":", "_")

    def disjoint_populations(self, other):
        # When a population is None we cannot draw conclusions
        #  - In Sonata there's no filtering and target may have multiple
        if self.population is None or other.population is None:
            return False
        # We are only sure if both are specified and different
        return self.population != other.population

    def overlap_byname(self, other):
        return not self.name or not other.name or self.name == other.name

    def overlap(self, other):
        """Are these target specs bound to overlap?
        If not, they still might be overlap, but Target gids need to be inspected
        """
        return self.population == other.population and self.overlap_byname(other)

    def __eq__(self, other):
        return other.population == self.population and other.name == self.name

    __hash__ = None


class TargetManager:
    def __init__(self, run_conf):
        """Initializes a new TargetManager"""
        self._run_conf = run_conf
        self._cell_manager = None
        self._targets = {}
        self._section_access = {}
        self._nodeset_reader = self._init_nodesets(run_conf)
        self._compartment_sets = self._init_compartment_sets(run_conf)
        # A list of the local node sets
        self.local_nodes = []

    @classmethod
    def _init_nodesets(cls, run_conf):
        config_nodeset_file = run_conf.get("config_node_sets_file", None)
        simulation_nodesets_file = run_conf.get("node_sets_file")
        if not simulation_nodesets_file and "TargetFile" in run_conf:
            target_file = run_conf["TargetFile"]
            if target_file.endswith(".json"):
                simulation_nodesets_file = target_file
        return (config_nodeset_file or simulation_nodesets_file) and NodeSetReader(
            config_nodeset_file, simulation_nodesets_file
        )

    @staticmethod
    def _init_compartment_sets(run_conf):
        return (
            libsonata.CompartmentSets.from_file(run_conf["compartment_sets_file"])
            if run_conf["compartment_sets_file"]
            else {}
        )

    def load_targets(self, circuit):
        """Provided that the circuit location is known and whether a nodes file has been
        specified, load any target files.
        Note that these will be moved into a TargetManager after the cells have been distributed,
        instantiated, and potentially split.
        """

        def _is_sonata_file(file_name):
            return file_name.endswith(".h5")

        nodes_file = circuit.get("CellLibraryFile")
        if nodes_file and _is_sonata_file(nodes_file) and self._nodeset_reader:
            self._nodeset_reader.register_node_file(nodes_file)

    @classmethod
    def create_global_target(cls):
        return NodesetTarget(TargetSpec.GLOBAL_TARGET_NAME, [])

    def register_cell_manager(self, cell_manager):
        self._cell_manager = cell_manager

    def register_target(self, target):
        self._targets[target.name] = target

    def register_local_nodes(self, local_nodes):
        """Registers the local nodes so that targets can be scoped to current rank"""
        self.local_nodes.append(local_nodes)

    def clear_simulation_data(self):
        self.local_nodes.clear()
        # deference SectionRefs to sections
        self._section_access.clear()

    def get_compartment_set(self, name: str):
        if self._compartment_sets is None or name not in self._compartment_sets:
            raise ConfigurationError(
                f"Unknown compartment set: {name}. "
                f"Known compartment sets: {', '.join(self._compartment_sets)}"
            )
        return self._compartment_sets[name]

    def get_target(self, target_spec: TargetSpec, target_pop=None):
        """Retrieves a target from any .target file or Sonata nodeset files.

        Targets are generic groups of cells not necessarily restricted to a population.
        When retrieved from the source files they can be cached.
        Targets retrieved from Sonata nodesets keep a reference to all Sonata
        node datasets and can be asked for a sub-target of a specific population.
        """
        if not isinstance(target_spec, TargetSpec):
            target_spec = TargetSpec(target_spec)
        if target_pop:
            target_spec.population = target_pop
        target_name = target_spec.name or TargetSpec.GLOBAL_TARGET_NAME
        target_pop = target_spec.population

        def get_concrete_target(target):
            """Get a more specific target, depending on specified population prefix"""
            target.update_local_nodes(self.local_nodes)
            return target if target_pop is None else target.make_subtarget(target_pop)

        # Check cached
        if target_name in self._targets:
            target = self._targets[target_name]
            return get_concrete_target(target)

        # Check if we can get a Nodeset
        target = self._nodeset_reader and self._nodeset_reader.read_nodeset(target_name)
        if target is not None:
            log_verbose("Retrieved `%s` from Sonata nodeset", target_spec)
            self.register_target(target)
            return get_concrete_target(target)

        raise ConfigurationError(f"Target {target_name} can't be loaded. Check target sources")

    @lru_cache  # noqa: B019
    def intersecting(self, target1, target2):
        """Checks whether two targets intersect"""
        target1_spec = TargetSpec(target1)
        target2_spec = TargetSpec(target2)
        if target1_spec.disjoint_populations(target2_spec):
            return False
        if target1_spec.overlap(target2_spec):
            return True

        # Couldn't get any conclusion from bare target spec
        # Obtain the targets to analyze
        t1, t2 = self.get_target(target1_spec), self.get_target(target2_spec)

        # Check for Sonata nodesets, they might have the same population and overlap
        if set(t1.populations) == set(t2.populations) and target1_spec.overlap_byname(target2_spec):
            return True

        # TODO: Investigate this might yield different results depending on the rank.
        return t1.intersects(t2)  # Otherwise go with full gid intersection

    def pathways_overlap(self, conn1, conn2, equal_only=False):
        src1, dst1 = conn1["Source"], conn1["Destination"]
        src2, dst2 = conn2["Source"], conn2["Destination"]
        if equal_only:
            return TargetSpec(src1) == TargetSpec(src2) and TargetSpec(dst1) == TargetSpec(dst2)
        return self.intersecting(src1, src2) and self.intersecting(dst1, dst2)

    def gid_to_sections(self, gid):
        """For a given gid, return a list of section references stored for random access.
        If the list does not exist, it is built and stored for future use.

        :param gid: GID of the cell
        :return: SerializedSections object with SectionRefs to every section in the cell
        """
        if gid not in self._section_access:
            cell_ref = self._cell_manager.get_cellref(gid)
            if cell_ref is None:
                logging.warning("No cell found for GID: %d", gid)
                return None
            result_serial = SerializedSections(cell_ref)
            self._section_access[gid] = result_serial
        else:
            result_serial = self._section_access[gid]

        return result_serial

    def location_to_point(self, gid, isec, ipt, offset):
        """Given a location for a cell, section id, segment id, and offset into the segment,
        create a list containing a section reference to there.

        :param gid: GID of the cell
        :param isec: Section index
        :param ipt: Distance to start of segment
        :param offset: Offset distance beyond the ipt (microns)
        :return: List with 1 item, where the synapse should go
        """
        # Soma connection, just zero it
        offset = max(offset, 0)

        result_point = TargetPointList(gid)
        cell_sections = self.gid_to_sections(gid)
        if not cell_sections:
            raise Exception("Getting locations for non-bg sims is not implemented yet...")

        if isec >= cell_sections.num_sections:
            raise Exception(
                f"Error: section {isec} out of bounds ({cell_sections.num_sections} "
                "total). Morphology section count is low, is this a good morphology?"
            )

        distance = 0.5
        tmp_section = cell_sections.isec2sec[int(isec)]

        if tmp_section is None:  # Assume we are in LoadBalance mode
            result_point.append(-1, None, -1)
        elif ipt == -1:
            # Sonata spec have a pre-calculated distance field.
            # In such cases, segment (ipt) is -1 and offset is that distance.
            offset = max(min(offset, 0.9999999), 0.0000001)
            result_point.append(int(isec), tmp_section, offset)
        else:
            # Adjust for section orientation and calculate distance
            section = tmp_section.sec
            if section.orientation() == 1:
                ipt = section.n3d() - 1 - ipt
                offset = -offset

            if ipt < section.n3d():
                distance = (section.arc3d(int(ipt)) + offset) / section.L
                distance = max(min(distance, 0.9999999), 0.0000001)

            if section.orientation() == 1:
                distance = 1 - distance

            result_point.append(int(isec), tmp_section, distance)
        return result_point


class NodeSetReader:
    """Implements reading Sonata Nodesets"""

    def __init__(self, config_nodeset_file, simulation_nodesets_file):
        def _load_nodesets_from_file(nodeset_file):
            if not nodeset_file:
                return libsonata.NodeSets("{}")
            return libsonata.NodeSets.from_file(nodeset_file)

        self._population_stores = {}
        self.nodesets = _load_nodesets_from_file(config_nodeset_file)
        simulation_nodesets = _load_nodesets_from_file(simulation_nodesets_file)
        duplicate_nodesets = self.nodesets.update(simulation_nodesets)
        if duplicate_nodesets:
            logging.warning("Some node set rules were replaced from %s", simulation_nodesets_file)

    def register_node_file(self, node_file):
        storage = libsonata.NodeStorage(node_file)
        for pop_name in storage.population_names:
            self._population_stores[pop_name] = storage

    def __contains__(self, nodeset_name):
        return nodeset_name in self.nodesets.names

    @property
    def names(self):
        return self.nodesets.names

    def read_nodeset(self, nodeset_name: str):
        """Build node sets capable of offsetting.
        The empty population has a special meaning in Sonata, it matches
        all populations in simulation
        """
        if nodeset_name not in self.nodesets.names:
            return None

        def _get_nodeset(pop_name):
            storage = self._population_stores.get(pop_name)
            population = storage.open_population(pop_name)
            # Create SelectionNodeSet object with 1-based gids
            try:
                node_selection = self.nodesets.materialize(nodeset_name, population)
            except libsonata.SonataError as e:
                msg = (
                    f"SonataError for nodeset {nodeset_name} "
                    f'from population "{pop_name}" : {e!s}, skip'
                )
                logging.warning(msg)
                return None
            if node_selection:
                logging.debug("Nodeset %s: Appending gids from %s", nodeset_name, pop_name)
                ns = SelectionNodeSet.from_zero_based_libsonata_selection(node_selection)
                ns.register_global(pop_name)
                return ns
            return None

        nodesets = (_get_nodeset(pop_name) for pop_name in self._population_stores)
        nodesets = [ns for ns in nodesets if ns]
        return NodesetTarget(nodeset_name, nodesets)


class NodesetTarget:
    """Represents a subset of nodes defined in a node_sets.json file.

    A `NodesetTarget` is constructed based on key-value pairs from the `node_sets.json` file,
    where each key defines an object containing a group of nodes. Nodes in a target can be
    selected from one or more populations, and they are internally organized into `nodesets`
    based on their respective populations.

    Example:
    Given populations:
    ```python
    populations = {
    "pop_A": [0, 1, 2],
    "pop_B": [1000, 1001],
    "pop_C": [2000, 2001, 2002]
    }
    ```
    A target could look like:
    ```python
    target = [0, 1, 1000, 1001]
    ```
    Internally, `NodesetTarget` would organize these nodes into:
    ```python
    nodesets = [
    SelectionNodeSet([0, 1]),
    SelectionNodeSet([1000, 1001])
    ]
    ```
    """

    def __init__(self, name, nodesets: list[SelectionNodeSet], local_nodes=None, **_kw):
        self.name = name
        self.nodesets = nodesets
        self.local_nodes = local_nodes

    def __repr__(self):
        nodesets_str = "\n  ".join(str(ns) for ns in self.nodesets)
        return f"NodesetTarget(name={self.name!r}, nodesets=[\n  {nodesets_str}\n])"

    def gid_count(self):
        """Total number of nodes"""
        return sum(len(ns) for ns in self.nodesets)

    def max_gid_count_per_population(self):
        """Max number of nodes per population

        Useful to distribute in multicycle runs
        """
        return max(len(ns) for ns in self.nodesets)

    def gids(self, raw_gids):
        """Return GIDs from a libsonata.Selection possibly made by a union of
        libsonata.Selections.
        """
        if not self.nodesets:
            logging.warning("Nodeset '%s' can't be materialized. No node populations", self.name)
            return np.array([])
        if raw_gids:
            if len(self.nodesets) > 1:
                raise TargetError(
                    "Can not get raw gids for Nodeset target with "
                    "multiple populations. "
                    f"{len(self.nodesets)} found."
                )
            return np.array(self.nodesets[0].gids(raw_gids=True))
        sel = libsonata.Selection([])
        for n in self.nodesets:
            sel |= n.selection(raw_gids=False)
        return sel.flatten()

    def __contains__(self, gid):
        """Determine if a given gid is included in the gid list for this target regardless of rank.

        Offsetting is taken into account
        """
        return self.contains(gid)

    def append_nodeset(self, nodeset: SelectionNodeSet):
        """Add a nodeset to the current target"""
        self.nodesets.append(nodeset)

    @property
    def population_names(self):
        return {ns.population_name for ns in self.nodesets}

    @property
    def populations(self):
        return {ns.population_name: ns for ns in self.nodesets}

    def make_subtarget(self, pop_name):
        """A nodeset subtarget contains only one given population"""
        nodesets = [ns for ns in self.nodesets if ns.population_name == pop_name]
        local_nodes = [n for n in self.local_nodes if n.population_name == pop_name]
        return NodesetTarget(f"{self.name}#{pop_name}", nodesets, local_nodes)

    def is_void(self):
        return len(self.nodesets) == 0

    def update_local_nodes(self, local_nodes):
        """Allows setting the local gids"""
        self.local_nodes = local_nodes

    def get_local_gids(self, raw_gids=False):
        """Return the list of target gids in this rank (with offset)"""
        assert self.local_nodes, "Local nodes not set"

        def pop_gid_intersect(nodeset: SelectionNodeSet, raw_gids):
            for local_ns in self.local_nodes:
                if local_ns.population_name == nodeset.population_name:
                    return nodeset.intersection(local_ns, raw_gids)
            return libsonata.Selection([])

        if raw_gids:
            assert len(self.nodesets) == 1, "Multiple populations when asking for raw gids"
            return pop_gid_intersect(self.nodesets[0], raw_gids=True).flatten()

        return reduce(
            or_,
            (pop_gid_intersect(ns, raw_gids=False) for ns in self.nodesets),
            libsonata.Selection([]),
        ).flatten()

    def get_point_list_from_compartment_set(
        self, cell_manager, compartment_set
    ) -> compat.List[TargetPointList]:
        """Builds a list of points grouped by GID from a compartment set,
        mapping sections and offsets for each relevant population.

        Note:
            The compartment locations in `compartment_set` are expected to be
            sorted and without duplicates. This is ensured by libsonata.
        """
        point_list = compat.List()
        population_name = compartment_set.population

        if population_name not in self.population_names:
            return point_list
        sel_node_set = self.populations[population_name]

        sel = libsonata.Selection(
            [(start - 1, stop - 1) for start, stop in sel_node_set.selection(raw_gids=True).ranges]
        )
        # compartment_set is 0-based
        for cl in compartment_set.filtered_iter(sel):
            raw_gid, section_id, offset = cl.node_id, cl.section_id, cl.offset
            # points are 1-based
            gid = sel_node_set._offset + raw_gid + 1
            cell = cell_manager.get_cell(gid)
            sec = cell.get_sec(section_id)
            if len(point_list) and point_list[-1].gid == gid:
                point_list[-1].append(section_id, Nd.SectionRef(sec), offset)
            else:
                point = TargetPointList(gid)
                point.append(section_id, Nd.SectionRef(sec), offset)
                point_list.append(point)

        return point_list

    def get_point_list(
        self,
        cell_manager,
        section_type: SectionType = SectionType.SOMA,
        compartment_type: CompartmentType = CompartmentType.CENTER,
    ) -> compat.List[TargetPointList]:
        """Retrieve TargetPointLists containing compartments (based on section type and
        compartment type) of any local cells on the cpu.

        Args:
            cell_manager: a cell manager or global cell manager
            section_type: SectionType
            compartment_type: CompartmentType

        Returns:
            list of TargetPointList containing the compartment position
            and retrieved section references
        """
        section_type_str = section_type.to_string()
        point_lists = compat.List()

        for gid in self.get_local_gids():
            point_list = TargetPointList(gid)
            cell = cell_manager.get_cell(gid)
            secs = getattr(cell.CellRef, section_type_str)

            for sec in secs:
                section_id = cell.get_section_id(sec)
                if compartment_type == CompartmentType.CENTER:
                    point_list.append(section_id, Nd.SectionRef(sec), 0.5)
                else:
                    for seg in sec:
                        point_list.append(section_id, Nd.SectionRef(sec), seg.x)
            point_lists.append(point_list)

        return point_lists

    def generate_subtargets(self, n_parts):
        """Generate sub NodeSetTarget per population for multi-cycle runs
        Returns:
            list of [sub_target_n_pop1, sub_target_n_pop2, ...]
        """
        if not n_parts or n_parts == 1:
            return False

        all_raw_gids = {ns.population_name: ns.gids(raw_gids=True) for ns in self.nodesets}

        new_targets = defaultdict(list)
        pop_names = list(all_raw_gids.keys())

        for cycle_i in range(n_parts):
            for pop in pop_names:
                # name sub target per populaton, to be registered later
                target_name = f"{pop}__{self.name}_{cycle_i}"
                target = NodesetTarget(target_name, [SelectionNodeSet().register_global(pop)])
                new_targets[pop].append(target)

        for pop, raw_gids in all_raw_gids.items():
            target_looper = itertools.cycle(new_targets[pop])
            for gid in raw_gids:
                target = next(target_looper)
                target.nodesets[0].add_gids([gid])

        # return list of subtargets lists of all pops per cycle
        return [
            [targets[cycle_i] for targets in new_targets.values()] for cycle_i in range(n_parts)
        ]

    def contains(self, items, raw_gids=False):
        """Return a bool or an array of bool's whether the elements are contained"""
        # Shortcut for empty target. Algorithm below would fail
        if not self.gid_count():
            return ([False] * len(items)) if hasattr(items, "__len__") else False

        gids = self.gids(raw_gids=True) if raw_gids else self.gids(raw_gids=False)
        contained = np.isin(items, gids, kind="table")
        return bool(contained) if contained.ndim == 0 else contained

    def intersects(self, other):
        """Check if two targets intersect. At least one common population has to intersect"""
        if self.population_names.isdisjoint(other.population_names):
            return False

        other_pops = other.populations  # may be created on the fly
        # We loop over one target populations and check the other existence and intersection
        for pop, nodeset in self.populations.items():
            if pop not in other_pops:
                continue
            if nodeset.intersects(other_pops[pop]):
                return True
        return False


class SerializedSections:
    """Serializes the sections of a cell for easier random access.
    Note that this is possible because the v field in the section has been assigned
    an integer corresponding to the target index as read from the morphology file.
    """

    def __init__(self, cell):
        self.num_sections = int(cell.nSecAll)
        # Initialize list to store SectionRef objects
        self.isec2sec = [None] * self.num_sections
        # Flag to control warning message display
        self._serialized_sections_warned = False

        for index, sec in enumerate(cell.all):
            # Accessing the 'v' value at location 0.0001 of the section
            v_value = sec(0.0001).v
            if v_value >= self.num_sections:
                logging.debug("{%s} v(1)={sec(1).v} n3d()={%f}", sec.name(), sec.n3d())
                raise Exception("Error: failure in mk2_isec2sec()")
            if v_value < 0:
                if not self._serialized_sections_warned:
                    logging.warning(
                        "SerializedSections: v(0.0001) < 0. index={%d} v()={%f}",
                        index,
                        v_value,
                    )
                    self._serialized_sections_warned = True
            else:
                # Store a SectionRef to the section at the index specified by v_value
                self.isec2sec[int(v_value)] = Nd.SectionRef(sec=sec)



"""A collection of Pure-Python MultiMaps"""

import collections.abc
from functools import reduce
from operator import add

import numpy as np


class GroupedMultiMap(collections.abc.Mapping):
    """A Multimap which groups values by key in a list."""

    __slots__ = (
        "_keys",  # stored as `np.array`
        "_values",  # stored as `list`
    )

    def __init__(self, np_keys, values):
        """Init GroupedMultiMap.

        Args:
            np_keys: The numpy array of the keys. Can be empty
            values: The array of the values, can be any indexable, but better if numpy
        """
        assert len(np_keys) == len(values), "Keys and values must have the same length"
        self._keys, self._values = self.sort_together(np_keys, values)
        self._keys, self._values = self._duplicates_to_list(self._keys, self._values)

    @staticmethod
    def sort_together(np_keys, values):
        sort_idxs = np_keys.argsort(kind="mergesort")  # need stability
        keys = np_keys[sort_idxs]
        if isinstance(values, np.ndarray):
            values = values[sort_idxs]
        else:
            values = [values[i] for i in sort_idxs]
        return keys, values

    def find(self, key):
        idx = np.searchsorted(self._keys, key)
        if idx == len(self._keys) or self._keys[idx] != key:
            return None
        return idx

    def keys(self):
        return self._keys

    def values(self):
        return self._values

    def __iter__(self):
        return iter(self._keys)

    def __len__(self):
        return len(self._keys)

    def __getitem__(self, key):
        idx = self.find(key)
        if idx is None:
            raise KeyError(f"{key} does not exist")
        return self._values[idx]

    def __setitem__(self, key, value):
        raise NotImplementedError(
            "Setitem is not allowed for performance reasons. "
            "Please build and add-inplace another MultiMap"
        )

    def items(self):
        return zip(self._keys, self._values)

    def __contains__(self, key):
        return self.find(key) is not None

    exists = __contains__  # Compat. w Hoc map

    @staticmethod
    def _concat(v1, v2):
        return (v1 if isinstance(v1, (list, tuple)) else list(v1)) + (
            v2 if isinstance(v2, (list, tuple)) else list(v2)
        )

    @staticmethod
    def _duplicates_to_list(np_keys, values):
        np_keys, indexes = np.unique(np_keys, return_index=True)
        if len(indexes) == 0:
            return np_keys, []
        beg_it = iter(indexes)
        end_it = iter(indexes)
        next(end_it)  # Discard first
        values = [values[next(beg_it) : end] for end in end_it] + [values[indexes[-1] :]]  # Last
        assert len(np_keys) == len(values)
        return np_keys, values

    def get(self, key, default=()):
        idx = self.find(key)
        if idx is None:
            return default
        return self._values[idx]

    def get_items(self, key):
        return self.get(key)

    def __iadd__(self, other):
        """Inplace add (incorporate other)"""
        self._keys, self._values = self.sort_together(
            np.concatenate((self._keys, other._keys)), self._concat(self._values, other._values)
        )
        self._keys, v_list = self._duplicates_to_list(self._keys, self._values)
        self._values = [reduce(add, subl) for subl in v_list]
        return self



from __future__ import annotations

import logging
from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import get_type_hints

from ._utils import run_only_rank0


@dataclass
class CoreSimulationConfig:
    # required
    outpath: str = field(metadata={"file_key": "outpath", "is_path": True})
    datpath: str = field(metadata={"file_key": "datpath", "is_path": True})
    tstop: float
    dt: float
    prcellgid: int
    celsius: float
    voltage: float
    cell_permute: int = field(metadata={"file_key": "cell-permute"})
    mpi: int

    # optional
    pattern: str = None
    seed: int = None
    model_stats: bool = field(default=None, metadata={"file_key": "model-stats"})
    report_conf: str = field(default=None, metadata={"file_key": "report-conf", "is_path": True})

    def __post_init__(self):
        # type coercion and path resolution
        hints = get_type_hints(self)
        for f in fields(self):
            val = getattr(self, f.name)
            if val is None:
                continue
            coerced = self._coerce_type(hints[f.name], val)
            if f.metadata.get("is_path", False) and coerced is not None:
                coerced = str(Path(coerced).resolve())
            setattr(self, f.name, coerced)

    @staticmethod
    def _coerce_type(typ, val):
        result = None

        if val is not None:
            if isinstance(val, typ):
                result = val
            elif typ is bool:
                result = val.strip().lower() in {"true", "1"} if isinstance(val, str) else bool(val)
            elif typ is int:
                result = int(val)
            elif typ is float:
                result = float(val)
            elif typ is str:
                result = str(val)
            else:
                result = val

        return result

    @run_only_rank0
    def dump(self, path: str | Path):
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        logging.info("Writing coreneuron simulation config file: %s", path.resolve())

        with path.open("w", encoding="utf-8") as fp:
            for f in fields(self):
                file_key = f.metadata.get("file_key", f.name)
                val = getattr(self, f.name)
                if file_key == "model-stats":
                    if val:
                        fp.write("model-stats\n")
                elif val is not None:
                    if isinstance(val, str):
                        fp.write(f"{file_key}='{val}'\n")
                    else:
                        fp.write(f"{file_key}={val}\n")

        logging.info("Done! coreneuron simulation config was written")

    @classmethod
    def load(cls, path: str | Path) -> CoreSimulationConfig:
        # precompute mapping once
        key_map = {f.metadata.get("file_key", f.name): f.name for f in fields(cls)}

        raw = {}
        with Path(path).open("r", encoding="utf-8") as fp:
            for line in fp:
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                if line == "model-stats":
                    raw[key_map["model-stats"]] = True
                    continue
                if "=" not in line:
                    raise ValueError(f"Malformed line in config: {line}")
                key, val = line.split("=", 1)
                key = key.strip()
                val = val.strip().strip("'").strip('"')

                if key not in key_map:
                    raise ValueError(f"Unknown config key: {key}")

                raw[key_map[key]] = val
        return cls(**raw)



import logging
from itertools import chain

from .connection import Connection, NetConType, ReplayMode
from .connection_manager import SynapseRuleManager
from .core.configuration import GlobalConfig
from .io.sonata_config import ConnectionTypes
from .io.synapse_reader import SonataReader, SynapseParameters
from .utils.logging import log_all


class NeuroModulationConnection(Connection):
    __slots__ = ("_neuromod_dtc", "_neuromod_strength")

    def __init__(
        self,
        sgid,
        tgid,
        src_pop_id=0,
        dst_pop_id=0,
        weight_factor=1.0,
        minis_spont_rate=None,
        configuration=None,
        mod_override=None,
        **kwargs,
    ):
        self._neuromod_strength = None
        self._neuromod_dtc = None
        super().__init__(
            sgid,
            tgid,
            src_pop_id,
            dst_pop_id,
            weight_factor,
            minis_spont_rate,
            configuration,
            mod_override,
            **kwargs,
        )

    neuromod_strength = property(
        lambda self: self._neuromod_strength,
        lambda self, val: setattr(self, "_neuromod_strength", val),
    )
    neuromod_dtc = property(
        lambda self: self._neuromod_dtc, lambda self, val: setattr(self, "_neuromod_dtc", val)
    )

    def finalize(
        self,
        cell,
        base_seed=0,
        *,
        replay_mode=ReplayMode.AS_REQUIRED,
        base_managers=None,
        **_kwargs,
    ):
        """Override the finalize process from the base class.
        NeuroModulatory events do not create synapses but link to existing cell synapses.
        A neuromodulatory connection from projections with match to the closest cell synapse.
        A spike coming from the neuromodulatory event (SynapseReplay) will trigger the
        NET_RECEIVE of the existing synapse, with the weight (binary 1/0), neuromod_strength,
        neuromod_dtc, and nc_type NC_MODULATOR
        """
        logging.debug("Finalize neuromodulation connection")

        self._netcons = []
        # Initialize member lists
        self._init_artificial_stims(cell, replay_mode)

        for syn_i, sec in self.sections_with_synapses:
            syn_params = self._synapse_params[syn_i]
            # We need to get all connections since we dont know the sgid
            # TODO: improve this by extracting all the relative distances only once
            base_conns = chain.from_iterable(
                base_manager.get_connections(self.tgid) for base_manager in base_managers
            )
            syn_obj = self._find_closest_cell_synapse(syn_params, base_conns)
            if syn_obj is None:
                logging.warning("No cell synapse associated to the neuromodulatory event")
                return 0
            if self._replay is not None:
                nc = self._replay.create_on(self, sec, syn_obj, syn_params)
                nc.weight[0] = int(self.weight_factor > 0)  # weight is binary 1/0, default 1
                nc.weight[1] = self.neuromod_strength or syn_params.neuromod_strength
                nc.weight[2] = self.neuromod_dtc or syn_params.neuromod_dtc
                self.netcon_set_type(nc, syn_obj, NetConType.NC_NEUROMODULATOR)
                if GlobalConfig.debug_conn == [self.tgid]:
                    log_all(
                        logging.DEBUG,
                        "Neuromodulatory event on tgid: %d, weights: [%f, %f, %f], nc_type: %d",
                        self.tgid,
                        nc.weight[0],
                        nc.weight[1],
                        nc.weight[2],
                        NetConType.NC_NEUROMODULATOR,
                    )

        return 1

    @staticmethod
    def _find_closest_cell_synapse(syn_params, base_conns):
        """Find the closest cell synapse by the location parameter"""
        if not base_conns:
            return None
        section_i = syn_params.isec
        location_i = syn_params.location
        min_diff = 0.05
        syn_obj = None
        for base_conn in base_conns:
            for syn_j, _ in base_conn.sections_with_synapses:
                params_j = base_conn._synapse_params[syn_j]
                if params_j.isec != section_i:
                    continue
                diff = abs(params_j.location - location_i)
                if diff < min_diff:
                    syn_obj = base_conn._synapses[syn_j]
                    min_diff = diff
        return syn_obj


class ModulationConnParameters(SynapseParameters):
    """Modulatory connection parameters.

    This class defines parameters for neuromodulatory connections by overriding
    `_fields` from `SynapseParameters`.

    Notes:
        - The field names are consistent with standard synapses for compatibility.
        - `location` is computed using the HOC function `TargetManager.locationToPoint`
          with `isec`, `offset`, and `ipt`.
        - `ipt` is set to -1 (not read from data) to ensure `locationToPoint` sets
          `location = offset`.
        - `weight` is used as a placeholder for replay stimulation; it defaults to 1.0
          and is later overwritten by the actual connection weight.

    The `_optional` and `_reserved` dictionaries are inherited from the base class.
    """

    _fields = {
        "sgid": "int64",
        "delay": "float64",
        "isec": "int64",
        "offset": "float64",
        "neuromod_strength": "float64",
        "neuromod_dtc": "float64",
        "ipt": "float64",
        "location": "float64",
        "weight": "float64",
    }


class NeuroModulationSynapseReader(SonataReader):
    Parameters = ModulationConnParameters
    custom_parameters = {"isec", "ipt", "offset", "weight"}

    def _load_params_custom(self, _populate, _read):
        super()._load_params_custom(_populate, _read)
        _populate("weight", 1)


class NeuroModulationManager(SynapseRuleManager):
    CONNECTIONS_TYPE = ConnectionTypes.NeuroModulation
    conn_factory = NeuroModulationConnection
    SynapseReader = NeuroModulationSynapseReader

    def _finalize_conns(self, tgid, conns, base_seed, **kwargs):
        """Override the function from the base class.
        Retrieve the base synapse connections with the same tgid.
        Pass the base connection managers (from all src populations except the neuromodulatory one)
        to the finalize process of superclass, to be processed by NeuroModulationConnection.
        """
        base_managers = [
            manager
            for src_pop, manager in self.cell_manager.connection_managers.items()
            if src_pop != self.src_node_population
        ]
        return super()._finalize_conns(
            tgid, conns, base_seed, base_managers=base_managers, **kwargs
        )



"""Module implementing interfaces to the several synapse readers (eg.: synapsetool, Hdf5Reader)"""

from __future__ import annotations

import logging

import libsonata
import numpy as np

from neurodamus.core import NeuronWrapper as Nd, ProgressBarRank0 as ProgressBar
from neurodamus.utils.logging import log_verbose
from neurodamus.utils.pyutils import gen_ranges


class SynapseParameters:
    """Synapse parameter names and dtypes for numpy recarrays
    following the SONATA specification.

    For detailed info on the parameters, see:
    https://sonata-extension.readthedocs.io/en/latest/sonata_tech.html#edge-file
    """

    _fields = {
        "sgid": np.int64,
        "delay": np.float64,
        "isec": np.int32,
        "ipt": np.int32,
        "offset": np.float64,
        "weight": np.float64,
        "U": np.float64,
        "D": np.float64,
        "F": np.float64,
        "DTC": np.float64,
        "synType": np.int32,
        "nrrp": np.int32,
        "u_hill_coefficient": np.float64,
        "conductance_ratio": np.float64,
        "maskValue": np.float64,
        "location": np.float64,
    }

    _optional = {"u_hill_coefficient": 0.0, "conductance_ratio": -1.0}
    _reserved = {"maskValue": -1.0, "location": 0.5}

    @classmethod
    def all_fields(cls):
        """Return all defined field names."""
        return set(cls._fields.keys())

    @classmethod
    def load_fields(cls):
        """Return all fields except reserved ones."""
        return cls.all_fields() - set(cls._reserved.keys())

    @classmethod
    def dtype(cls, extra_fields=None):
        """Return dtype including optional extra fields (all float64)."""
        fields = list(cls._fields.items())

        if extra_fields is not None:
            fields.extend((name, np.float64) for name in extra_fields)

        return np.dtype(fields)

    @classmethod
    def fields(cls, exclude: set = (), with_translation: dict | None = None):
        """Return list of fields with optional exclusion and translation.

        Returns list of tuples (field_name, translated_name or is_optional).
        """
        optional_keys = cls._optional.keys()
        fields = cls.load_fields() - exclude if exclude else cls.load_fields()

        if with_translation:
            return [(f, with_translation.get(f, f), f in optional_keys) for f in fields]
        return [(f, f in optional_keys) for f in fields]

    @staticmethod
    def _patch_delay_fp_inaccuracies(records):
        """Round 'delay' to nearest multiple of Nd.dt to fix fp inaccuracies."""
        if len(records) == 0 or "delay" not in records.dtype.names:
            return
        dt = Nd.dt
        records.delay = (records.delay / dt + 1e-5).astype("i4") * dt

    @staticmethod
    def _constrained_hill(K_half, y):  # noqa: N803
        """Constrained Hill function for scaling synaptic parameters.

        Note: it is iused only in scale_U_param. It is its own function
        because it deserves to be tested separately.
        """
        K4 = K_half**4
        y4 = y**4
        return (K4 + 16) / 16 * y4 / (K4 + y4)

    @staticmethod
    def _patch_scale_U_param(syn_params, extra_cellular_calcium, extra_scale_vars):
        """Scale 'U' and other vars using constrained Hill function based on
        extracellular calcium.
        """
        if len(syn_params) == 0 or extra_cellular_calcium is None:
            return

        scale_factors = SynapseParameters._constrained_hill(
            syn_params.u_hill_coefficient, extra_cellular_calcium
        )
        syn_params.U *= scale_factors
        for var in extra_scale_vars:
            syn_params[var] *= scale_factors

    @classmethod
    def make_synapse_parameters_array(
        cls,
        data: dict,
        extra_fields: list[str],
        extra_cellular_calcium: float | None,
        extra_scale_vars: list[str],
    ):
        """Create a recarray from data with optional extra fields and apply patches."""
        if not data:
            return np.recarray(0, dtype=cls.dtype(extra_fields=None))
        edge_count = len(next(iter(data.values())))
        arr = np.recarray(edge_count, dtype=cls.dtype(extra_fields=extra_fields))

        for name in arr.dtype.names:
            if name in cls._reserved:
                arr[name] = cls._reserved[name]
            elif name in cls._optional:
                arr[name] = data.get(name, cls._optional[name])
            elif name in data:
                arr[name] = data[name]
            else:
                raise AttributeError(f"Missing mandatory attribute {name} in the SONATA edge file")

        cls._patch_delay_fp_inaccuracies(arr)
        cls._patch_scale_U_param(arr, extra_cellular_calcium, extra_scale_vars)

        return arr


class SonataReader:
    """Reader for SONATA edge files.

    Uses libsonata directly and contains a bunch of workarounds to accomodate files
    created in the transition to SONATA. Also translates all GIDs from 0-based as on disk
    to the 1-based convention in Neurodamus.

    Will read each attribute for multiple GIDs at once and cache read data in a columnar
    fashion.

    FIXME Remove the caching at the np.recarray level.
    """

    SYNAPSE_INDEX_NAMES = ("synapse_index",)
    LOOKUP_BY_TARGET_IDS = True  # False to lookup by Source Ids
    # SynapseParameters knows how to load the parameters
    # Child classes of SonataReader can override this with another class
    # probably inherited from SynapseParameters to load a different set of parameters
    Parameters = SynapseParameters  # By default we load synapses
    EMPTY_DATA = {}

    custom_parameters = {"isec", "ipt", "offset"}
    """Custom parameters are skipped from direct loading and trigger _load_params_custom()"""

    parameter_mapping = {
        "weight": "conductance",
        "U": "u_syn",
        "D": "depression_time",
        "F": "facilitation_time",
        "DTC": "decay_time",
        "synType": "syn_type_id",
        "nrrp": "n_rrp_vesicles",
        "conductance_ratio": "conductance_scale_factor",
    }

    def __init__(self, edge_file, population=None, *_, **kw):
        self._ca_concentration = kw.get("extracellular_calcium")
        self._syn_params = {}  # Parameters cache by post-gid (previously loadedMap)
        self._open_file(edge_file, population, kw.get("verbose", False))
        # NOTE u_hill_coefficient and conductance_scale_factor are optional, BUT
        # while u_hill_coefficient can always be readif avail, conductance reader may not.
        self._uhill_property_avail = self.has_property("u_hill_coefficient")
        self._extra_fields = set()
        self._extra_scale_vars = []

    def configure_override(self, mod_override):
        if not mod_override:
            return

        override_helper = mod_override + "Helper"
        Nd.load_hoc(override_helper)

        # Read attribute names with format "attr1;attr2;attr3"
        attr_names = getattr(Nd, override_helper + "_NeededAttributes", None)
        if attr_names:
            log_verbose(
                'Reading parameters "{}" for mod override: {}'.format(
                    ", ".join(attr_names.split(";")), mod_override
                )
            )
            self._extra_fields = set(attr_names.split(";"))

        # Read attribute names with format "attr1;attr2;attr3"
        attr_names = getattr(Nd, override_helper + "_UHillScaleVariables", None)
        if attr_names:
            self._extra_scale_vars = attr_names.split(";")

    def get_synapse_parameters(self, gid) -> np.recarray:
        """Return the synapse parameters record array for the given gid,
        loading and caching it if needed.
        """
        syn_params = self._syn_params.get(gid)
        if syn_params is None:
            # prepare data for the gid
            data = self._data.get(gid)
            if data is None:  # not in _data
                self._preload_data_chunk([gid])
                data = self._data[gid]

            # create the synapse parameters array, already patched
            syn_params = self.Parameters.make_synapse_parameters_array(
                data, self._extra_fields, self._ca_concentration, self._extra_scale_vars
            )
            # cache the results
            self._syn_params[gid] = syn_params
        return syn_params

    def _open_file(self, src, population, _):
        """Initializes the reader, opens the synapse file"""
        try:
            from mpi4py import MPI

            hdf5_reader = libsonata.make_collective_reader(
                MPI.COMM_WORLD, collective_metadata=False, collective_transfer=True
            )
        except ModuleNotFoundError:
            hdf5_reader = libsonata.Hdf5Reader()

        storage = libsonata.EdgeStorage(src, hdf5_reader=hdf5_reader)
        if not population:
            assert len(storage.population_names) == 1, f"Populations: {storage.population_names}"
            population = next(iter(storage.population_names))
        self._population = storage.open_population(population)
        # A cache which stores all the fields for each gid. E.g. {1: {"sgid": property_numpy}}
        self._data = {}
        # A cache for connection counts, used mostly in dry run
        self._counts = {}

    def has_property(self, field_name):
        """Checks whether source data has the given additional field."""
        if field_name in self.SYNAPSE_INDEX_NAMES:
            return True
        return field_name in self._population.attribute_names

    def get_property(self, gid, field_name):
        """Retrieves a full pre-loaded property given a gid and the property name."""
        return self._data[gid][field_name]

    def preload_data(self, gids, minimal_mode=False):
        """Preload SONATA fields for the specified IDs.
        Set minimal_mode to True to read a single synapse per connection
        """
        # TODO: limit the number of cells per chunk in production.
        #       Ensuring the number of chunks must be the same in all ranks (collective)!
        CHUNK_SIZE = 1000
        if not minimal_mode or len(gids) < CHUNK_SIZE:
            return

        ranges = list(gen_ranges(len(gids), CHUNK_SIZE))
        for start, end in ProgressBar.iter(ranges, name="Prefetching"):
            self._preload_data_chunk(gids[start:end], minimal_mode)

    def _preload_data_chunk(self, gids, minimal_mode=False):  # noqa: C901
        """Preload all synapses for a number of gids, respecting Parameters and _extra_fields"""
        # NOTE: to disambiguate, gids are 1-based cell ids, while node_ids are 0-based sonata ids
        compute_fields = {"sgid", "tgid", *self.SYNAPSE_INDEX_NAMES}
        orig_needed_gids_set = set(gids) - set(self._data.keys())
        needed_gids = sorted(orig_needed_gids_set)

        def get_edge_and_lookup_gids(needed_gids: libsonata.Selection):
            """Retrieve edge and corresponding gid for"""
            node_ids = np.array(needed_gids, dtype="int64") - 1
            if self.LOOKUP_BY_TARGET_IDS:
                edge_ids = self._population.afferent_edges(node_ids)
                return edge_ids, self._population.target_nodes(edge_ids) + 1
            edge_ids = self._population.efferent_edges(node_ids)
            return edge_ids, self._population.source_nodes(edge_ids) + 1

        # NOTE: needed_edge_ids, lookup_gids are used in _populate and _read
        needed_edge_ids, lookup_gids = get_edge_and_lookup_gids(needed_gids)

        # Find and exclude gids without data
        different_gids_edge_i = np.diff(lookup_gids, prepend=np.nan).nonzero()[0]
        needed_gids = sorted(lookup_gids[different_gids_edge_i])
        for gid in orig_needed_gids_set - set(needed_gids):
            self._data.setdefault(gid, self.EMPTY_DATA)

        # In minimal mode read a single synapse (the first) of each target gid
        if minimal_mode:
            needed_edge_ids = libsonata.Selection(needed_edge_ids.flatten()[different_gids_edge_i])
            lookup_gids = lookup_gids[different_gids_edge_i]

        def _populate(field, data):
            # Populate cache. Unavailable entries are stored as a plain -1
            if data is None:
                data = -1
            for gid in needed_gids:
                existing_gid_data = self._data.setdefault(gid, {})
                existing_gid_data[field] = data if np.isscalar(data) else data[lookup_gids == gid]

        def _read(attribute, optional=False):
            if attribute in self._population.attribute_names:
                return self._population.get_attribute(attribute, needed_edge_ids)
            if optional:
                log_verbose("Defaulting to -1.0 for attribute %s", attribute)
                return -1
            raise AttributeError(f"Missing attribute {attribute} in the SONATA edge file")

        # Populate the opposite node id
        if self.LOOKUP_BY_TARGET_IDS:
            _populate("sgid", self._population.source_nodes(needed_edge_ids) + 1)
        else:
            _populate("tgid", self._population.target_nodes(needed_edge_ids) + 1)

        # Make synapse index in the file explicit
        for name in sorted(self.SYNAPSE_INDEX_NAMES):
            _populate(name, needed_edge_ids.flatten())

        # Generic synapse parameters
        fields_load_sonata = self.Parameters.fields(
            exclude=self.custom_parameters | compute_fields, with_translation=self.parameter_mapping
        )
        for field, sonata_attr, is_optional in sorted(fields_load_sonata):
            _populate(field, _read(sonata_attr, is_optional))

        if self.custom_parameters:
            if minimal_mode:
                _populate("isec", 0)
                _populate("ipt", -1)
                _populate("offset", 0)
                return  # done! Skip extra fields
            self._load_params_custom(_populate, _read)

        # Extend Gids data with the additional requested fields
        # This has to work for when we call preload() a second/third time
        # so we are unsure about which gids were loaded what properties
        # We nevertheless can skip any base fields
        extra_fields = self._extra_fields - (self.Parameters.all_fields() | compute_fields)
        for field in sorted(extra_fields):
            now_needed_gids = sorted(
                {
                    gid
                    for gid in gids
                    if (data := self._data[gid]) is not self.EMPTY_DATA and field not in data
                }
            )
            if needed_gids != now_needed_gids:
                needed_gids = now_needed_gids
                needed_edge_ids, lookup_gids = get_edge_and_lookup_gids(needed_gids)
            sonata_attr = self.parameter_mapping.get(field, field)
            _populate(field, _read(sonata_attr))

    def _load_params_custom(self, _populate, _read):
        # Position of the synapse
        if self.has_property("afferent_section_id"):
            _populate("isec", _read("afferent_section_id"))
            # SONATA compliant synapse position: (section, section_fraction) takes precedence
            # over the older (section, segment, segment_offset) synapse position.
            #
            # Re-using field names for historical reason, where -1 means N/A.
            # FIXME Use dedicated fields
            if self.has_property("afferent_section_pos"):
                _populate("ipt", -1)
                _populate("offset", _read("afferent_section_pos"))
            # This was a temporary naming scheme
            # FIXME Circuits using this field should be fixed
            elif self.has_property("afferent_section_fraction"):
                logging.warning(
                    "Circuit uses non-standard compliant attribute `afferent_section_fraction`"
                )
                _populate("ipt", -1)
                _populate("offset", _read("afferent_section_fraction"))
            else:
                logging.warning(
                    "Circuit is missing standard compliant attribute `afferent_section_pos`"
                )
                _populate("ipt", _read("afferent_segment_id"))
                _populate("offset", _read("afferent_segment_offset"))
        else:
            # FIXME All this should go the way of the dodo
            logging.warning(
                "Circuit uses attribute notation using `morpho_` and is not SONATA compliant"
            )
            _populate("isec", _read("morpho_section_id_post"))
            if self.has_property("morpho_section_fraction_post"):
                _populate("ipt", -1)
                _populate("offset", _read("morpho_section_fraction_post"))
            else:
                _populate("ipt", _read("morpho_segment_id_post"))
                _populate("offset", _read("morpho_offset_segment_post"))

    def get_counts(self, tgids):
        """Counts synapses for the given target neuron ids. Returns a dict"""
        node_ids = tgids - 1
        edge_ids = self._population.afferent_edges(node_ids)
        target_nodes = self._population.target_nodes(edge_ids)
        unique_nodes, counts = np.unique(target_nodes, return_counts=True)
        unique_gids = unique_nodes + 1
        counts_dict = dict(zip(unique_gids, counts))
        for gid in tgids:
            counts_dict.setdefault(gid, 0)
        return counts_dict

    def get_conn_counts(self, tgids):
        """Counts synapses per connetion for all the given target neuron ids.
        Returns a dict whose value is a numpy stuctured array
        """
        if missing_gids := set(tgids) - set(self._counts):
            missing_gids = np.fromiter(missing_gids, dtype="uint32")
            missing_gids.sort()
            missing_nodes = missing_gids - 1
            edge_ids = self._population.afferent_edges(missing_nodes)
            target_nodes = self._population.target_nodes(edge_ids)
            source_nodes = self._population.source_nodes(edge_ids)
            connections = np.empty(len(target_nodes), dtype="uint64,uint64")
            connections["f0"] = target_nodes + 1  # nodes to 1-based gids
            connections["f1"] = source_nodes + 1

            tgt_src_pairs, counts = np.unique(connections, return_counts=True)
            pairs_start_i = np.diff(tgt_src_pairs["f0"], prepend=np.nan, append=np.nan).nonzero()[0]

            for conn_i, start_i in enumerate(pairs_start_i[:-1]):
                end_i = pairs_start_i[conn_i + 1]
                tgid = tgt_src_pairs["f0"][start_i]
                tgid_counts = {tgt_src_pairs["f1"][j]: counts[j] for j in range(start_i, end_i)}
                self._counts[tgid] = tgid_counts

        return {tgid: self._counts.get(tgid, self.EMPTY_DATA) for tgid in tgids}



"""Collection of utility functions related to clearing the used memory in neurodamus-py or NEURON"""

import ctypes
import ctypes.util
import gzip
import heapq
import json
import logging
import math
import multiprocessing
import os
import pickle  # noqa: S403
from collections import Counter, defaultdict
from pathlib import Path

import numpy as np
import psutil

from .compat import Vector
from neurodamus.core import MPI, NeuronWrapper as Nd, run_only_rank0
from neurodamus.io.sonata_config import ConnectionTypes

# The factor to multiply the cell + synapses memory usage by to get the simulation memory estimate.
# This is an heuristic estimate based on tests on multiple circuits.
# More info in docs/architecture.rst.
SIM_ESTIMATE_FACTOR = 2.5
MAX_ALLOCATION_STEPS = 10  # Maximum number of steps to reduce the batch size in distribute_cells


def trim_memory():
    """malloc_trim - release free memory from the heap (back to the OS)

    * We should only run malloc_trim if we are using the default glibc memory allocator.
    When using a custom allocator such as jemalloc, this could cause unexpected behavior
    including segfaults.

    * The malloc_trim function returns 1 if memory was actually
    released back to the system, or 0 if it was not possible to
    release any memory.
    """
    if os.getenv("LD_PRELOAD") and "jemalloc" in os.getenv("LD_PRELOAD"):
        logging.warning(
            "malloc_trim works only with the default glibc memory allocator. "
            "Please avoid using jemalloc (for now, we skip malloc_trim)."
        )
        logging.info("malloc_trim: not possible to release any memory.")
    else:
        try:
            path_libc = ctypes.util.find_library("c")
            libc = ctypes.CDLL(path_libc)
            memory_trimmed = libc.malloc_trim(0)
            if memory_trimmed:
                logging.info("malloc_trim: memory released back to the system.")
        except (OSError, AttributeError):
            logging.exception("Unable to call malloc_trim.")
            logging.info("malloc_trim: not possible to release any memory.")


def pool_shrink():
    """Shrink NEURON ArrayPools to the exact size still being used by NEURON.
    This is useful when clearing all the Nodes from NEURON (cells and synapses) and the underlying
    data structures holding the data of their mechanisms need to be shrinked down since they are
    not used any more.
    This feature is enabled in NEURON after version 8.2.2a.
    See https://github.com/neuronsimulator/nrn/pull/2033 for more information.
    """
    try:
        cv = Nd.CVode()
        cv.poolshrink(1)
    except AttributeError:
        logging.warning(
            "Cannot shrink NEURON ArrayPools. NEURON does not support CVode().poolshrink(1)"
        )


def free_event_queues():
    """Apart from the NEURON SelfEventQueue and TQItem pools being clear we also free them from any
    memory still referencing to.
    """
    try:
        cv = Nd.CVode()
        cv.free_event_queues()
    except AttributeError:
        logging.warning(
            "Cannot clear NEURON event queues.NEURON does not support CVode().free_event_queues()"
        )


def print_node_level_mem_usage():
    """Print statistics of the memory usage per compute node."""
    from neurodamus.core._shmutils import SHMUtil

    # The association of MPI ranks to compute nodes isn't
    # always know. If unknown, don't print anything.
    if not SHMUtil.is_node_id_known():
        return

    rss = SHMUtil.get_nodewise_rss()

    min_usage_mb = np.min(rss) / 2**20
    max_usage_mb = np.max(rss) / 2**20
    avg_usage_mb = np.mean(rss) / 2**20
    dev_usage_mb = np.std(rss) / 2**20

    logging.info(
        "Memusage (RSS) per node [MB]: Max=%.2lf, Min=%.2lf, Mean(Stdev)=%.2lf(%.2lf)",
        max_usage_mb,
        min_usage_mb,
        avg_usage_mb,
        dev_usage_mb,
    )


def get_task_level_mem_usage():
    """Return statistics of the memory usage per MPI task."""
    usage_mb = get_mem_usage_kb() / 1024

    min_usage_mb = MPI.pc.allreduce(usage_mb, MPI.MIN)
    max_usage_mb = MPI.pc.allreduce(usage_mb, MPI.MAX)
    avg_usage_mb = MPI.pc.allreduce(usage_mb, MPI.SUM) / MPI.size

    dev_usage_mb = math.sqrt(MPI.pc.allreduce((usage_mb - avg_usage_mb) ** 2, MPI.SUM) / MPI.size)

    return min_usage_mb, max_usage_mb, avg_usage_mb, dev_usage_mb


def print_task_level_mem_usage():
    """Print statistics of the memory usage per MPI task."""
    min_usage_mb, max_usage_mb, avg_usage_mb, dev_usage_mb = get_task_level_mem_usage()

    logging.info(
        "Memusage (RSS) per task [MB]: Max=%.2lf, Min=%.2lf, Mean(Stdev)=%.2lf(%.2lf)",
        max_usage_mb,
        min_usage_mb,
        avg_usage_mb,
        dev_usage_mb,
    )


def print_mem_usage():
    """Print memory usage information across all ranks."""
    print_node_level_mem_usage()
    print_task_level_mem_usage()


def get_mem_usage_kb():
    """Return memory usage information across all ranks, in KiloBytes"""
    usage_kb = psutil.Process(os.getpid()).memory_info().rss / 1024

    return usage_kb


def pretty_printing_memory_mb(memory_mb):
    """A simple function that given a memory usage in MB
    returns a string with the most appropriate unit.
    """
    if memory_mb < 1024:
        return f"{memory_mb:.2f} MB"
    if memory_mb < 1024**2:
        return "%.2lf GB" % (memory_mb / 1024)
    if memory_mb < 1024**3:
        return "%.2lf TB" % (memory_mb / 1024**2)
    return "%.2lf PB" % (memory_mb / 1024**3)


@run_only_rank0
def print_allocation_stats(rank_memory):
    """Print statistics of the memory allocation across ranks.

    Args:
        rank_memory (dict): A dictionary where keys are rank IDs
                            and values are the total memory load on each rank.
    """
    logging.debug("Total memory per rank/cycle: %s", rank_memory)
    import statistics

    for pop, rank_dict in rank_memory.items():
        values = list(rank_dict.values())
        logging.info("Population: %s", pop)
        logging.info("Mean allocation per rank [KB]: %s", round(statistics.mean(values)))
        try:
            stdev = round(statistics.stdev(values))
        except statistics.StatisticsError:
            stdev = 0
        logging.info("Stdev of allocation per rank [KB]: %s", stdev)


@run_only_rank0
def export_allocation_stats(rank_allocation, filename, ranks, cycles=1):
    """Export allocation dictionary to a serialized pickle file."""
    compressed_data = gzip.compress(pickle.dumps(rank_allocation))
    Path(f"{filename}_r{ranks}_c{cycles}.pkl.gz").write_bytes(compressed_data)


class SynapseMemoryUsage:
    """A small class that works as a lookup table
    for the memory used by each type of synapse.
    The values are in KB. The values cannot be set by the user.
    """

    _synapse_memory_usage = {
        ConnectionTypes.Synaptic: 1.9,  # Average between 1.7 (AMPA) and 2.0 (GABAAB)
        ConnectionTypes.GapJunction: 2.0,
    }

    @classmethod
    def get_memory_usage(cls, count, synapse_type=ConnectionTypes.Synaptic):
        return count * cls._synapse_memory_usage[synapse_type]


class DryRunStats:
    _MEMORY_USAGE_FILENAME = "cell_memory_usage.json"
    _ALLOCATION_FILENAME = "allocation"

    _alloc_cache = None

    @staticmethod
    def defaultdict_vector():
        return defaultdict(Vector)

    @staticmethod
    def defaultdict_float():
        return defaultdict(float)

    def __init__(self) -> None:
        self.metype_memory = {}
        self.metype_cell_syn_average = Counter()
        self.pop_metype_gids = {}
        self.metype_counts = Counter()
        self.synapse_counts = defaultdict(int)  # [syn_type -> count]
        self.suggested_nodes = 0
        self.synapse_memory_total = 0
        _, _, self.base_memory, _ = get_task_level_mem_usage()

    def import_allocation_stats(self, filename, cycle_i=0, ignore_cache=False) -> dict:
        """Import allocation dictionary from serialized pickle file."""

        def convert_to_standard_types(obj):
            """Converts an object containing defaultdicts of Vectors to standard Python types."""
            result = {}
            for population, vectors in obj.items():
                result[population] = {
                    key: np.array(vector) for key, vector in vectors.items() if key[0] == MPI.rank
                }
            return result

        if self._alloc_cache is None or ignore_cache:
            logging.warning("Loading allocation stats from %s...", filename)
            with gzip.open(filename, "rb") as f:
                data = pickle.load(f)  # noqa: S301
            DryRunStats._alloc_cache = convert_to_standard_types(data)
        else:
            logging.warning("Using cached allocation stats.")

        # Filter the cached data based on the cycle index
        filtered_alloc = {}
        for population, vectors in self._alloc_cache.items():
            filtered_alloc[population] = {
                key: vector for key, vector in vectors.items() if key[1] == cycle_i
            }

        return filtered_alloc

    @run_only_rank0
    def estimate_cell_memory(self) -> float:
        from .logging import log_verbose

        memory_total = 0
        log_verbose("+{:=^81}+".format(" METype Memory Estimates (MiB) "))
        log_verbose(
            "| {:^40s} | {:^10s} | {:^10s} | {:^10s} |".format(
                "METype", "Mem/cell", "N Cells", "Mem Total"
            )
        )
        log_verbose("+{:-^81}+".format(""))
        for metype, count in self.metype_counts.items():
            metype_mem = self.metype_memory[metype] / 1024
            metype_total = count * metype_mem
            memory_total += metype_total
            log_verbose(
                f"| {metype:<40s} | {metype_mem:10.2f} | {count:10.0f} | {metype_total:10.1f} |"
            )
        log_verbose("+{:-^81}+".format(""))
        self.cell_memory_total = memory_total
        logging.info("  Total memory usage for cells: %s", pretty_printing_memory_mb(memory_total))
        return memory_total

    def add(self, other):
        self.metype_memory.update(other.metype_memory)
        self.metype_counts += other.metype_counts

    def collect_all_mpi(self):
        # We combine memory dict via update(). That means if a previous circuit computed
        # cells for the same METype (hopefully unlikely!) the last estimate prevails.
        self.metype_memory = MPI.py_reduce(self.metype_memory, {}, lambda x, y: x.update(y))
        self.metype_cell_syn_average = MPI.py_sum(self.metype_cell_syn_average, Counter())
        self.metype_counts = self.metype_counts  # Cell counts is complete in every rank

    @run_only_rank0
    def export_cell_memory_usage(self):
        with open(self._MEMORY_USAGE_FILENAME, "w", encoding="utf-8") as fp:
            json.dump(self.metype_memory, fp, sort_keys=True, indent=4)

    def try_import_cell_memory_usage(self):
        if not os.path.exists(self._MEMORY_USAGE_FILENAME):
            return
        logging.info("Loading memory usage from %s...", self._MEMORY_USAGE_FILENAME)
        with open(self._MEMORY_USAGE_FILENAME, encoding="utf-8") as fp:
            self.metype_memory = json.load(fp)

    def collect_display_syn_counts(self):
        from .logging import log_verbose

        master_counter = MPI.py_sum(self.synapse_counts, Counter())

        # Done with MPI. Use rank0 to display
        if MPI.rank != 0:
            return

        logging.info(" - Estimated synapse memory usage (MB):")
        log_verbose("+{:=^68}+".format(" Synapse Count "))
        log_verbose("| {:^40s} | {:^10s} | {:^10s} |".format("Synapse Type", "Count", "Memory"))
        log_verbose("+{:-^68}+".format(""))

        # Some synapse types are numeric, others are strings, so we need to handle both
        for syn_type, count in master_counter.items():
            mem_mb = SynapseMemoryUsage.get_memory_usage(count, syn_type) / 1024
            self.synapse_memory_total += mem_mb
            mem_str = pretty_printing_memory_mb(mem_mb)
            log_verbose(f"| {syn_type!s:<40s} | {count:10.0f} | {mem_str:>10s} |")

        log_verbose("+{:-^68}+".format(""))  # Close table
        logging.info(" - TOTAL : %s", pretty_printing_memory_mb(self.synapse_memory_total))

    @run_only_rank0
    def display_total(self):
        logging.info("+{:=^57}+".format(" Total Memory Estimates "))
        logging.info("| {:^40s} | {:^12s} |".format("Item", "Memory (MiB)"))
        logging.info("+{:-^57}+".format(""))
        full_overhead = self.base_memory * MPI.size
        logging.info("| {:<40s} | {:12.1f} |".format(f"Overhead (ranks={MPI.size})", full_overhead))
        logging.info("| {:<40s} | {:12.1f} |".format("Cells", self.cell_memory_total))
        logging.info("| {:<40s} | {:12.1f} |".format("Synapses", self.synapse_memory_total))
        self.simulation_estimate = (
            self.cell_memory_total + self.synapse_memory_total
        ) * SIM_ESTIMATE_FACTOR
        logging.info("| {:<40s} | {:12.1f} |".format("Simulation", self.simulation_estimate))
        logging.info("+{:-^57}+".format(""))
        grand_total = (
            full_overhead
            + self.cell_memory_total
            + self.synapse_memory_total
            + self.simulation_estimate
        )
        grand_total = pretty_printing_memory_mb(grand_total)
        logging.info("| {:<40s} | {:>12s} |".format("GRAND TOTAL", grand_total))
        logging.info("+{:-^57}+".format(""))

    def total_memory_available():
        """Returns the total memory available in the system in MB"""
        try:
            virtual_memory = psutil.virtual_memory()
            return virtual_memory.total / (1024 * 1024)  # Total available memory in MB
        except Exception:
            logging.exception("Error")
            return None

    @run_only_rank0
    def suggest_nodes(self, margin):
        """A function to calculate the suggested number of nodes to run the simulation
        The function takes into account the fact that the memory overhead is
        variable with the amount of ranks the simulation it's ran with.
        One can also specify a custom margin to add to the memory usage.
        """
        try:
            ranks_per_node = os.cpu_count()
        except AttributeError:
            ranks_per_node = multiprocessing.cpu_count()

        full_overhead = self.base_memory * ranks_per_node

        # initialize variable for iteration
        est_nodes = 0
        prev_est_nodes = None
        max_iter = 5
        iter_count = 0

        while (prev_est_nodes is None or est_nodes != prev_est_nodes) and iter_count < max_iter:
            prev_est_nodes = est_nodes
            simulation_estimate = (
                self.cell_memory_total + self.synapse_memory_total
            ) * SIM_ESTIMATE_FACTOR
            mem_usage_per_node = (
                full_overhead
                + self.cell_memory_total
                + self.synapse_memory_total
                + simulation_estimate
            )
            mem_usage_with_margin = mem_usage_per_node * (1 + margin)
            est_nodes = math.ceil(mem_usage_with_margin / DryRunStats.total_memory_available())
            full_overhead = self.base_memory * ranks_per_node * est_nodes
            iter_count += 1

        return est_nodes

    @run_only_rank0
    def display_node_suggestions(self):
        """Display suggestions for how many nodes are approximately
        necessary to run the simulation based on the memory available
        on the current node.
        """
        node_total_memory = DryRunStats.total_memory_available()
        if node_total_memory is None:
            logging.warning("Unable to get the total memory available on the current node.")
            return
        self.suggested_nodes = self.suggest_nodes(0.3)
        logging.info(
            "Based on the memory available on the current node, "
            "it is suggested to use at least %s node(s).",
            self.suggested_nodes,
        )
        logging.info(
            "This is just a suggestion and the actual number of nodes "
            "needed to run the simulation may be different."
        )
        logging.info(
            "The calculation was based on a total memory available of %s on the current node.",
            pretty_printing_memory_mb(node_total_memory),
        )
        logging.info(
            "Please remember that it is suggested to use the same class of nodes "
            "for both the dryrun and the actual simulation."
        )

    def get_num_target_ranks(self, num_ranks):
        """Return the number of ranks to target for dry-run load balancing"""
        if num_ranks is None:
            logging.info("No number of ranks specified. Using suggested number of nodes.")
            logging.info("Detected number of physical cores: %d", psutil.cpu_count(logical=False))
            return self.suggested_nodes * psutil.cpu_count(logical=False)
        return int(num_ranks)

    @run_only_rank0
    def distribute_cells(
        self, num_ranks: int, cycles: int = 1, batch_size=10
    ) -> tuple[dict, dict, dict]:
        """Distributes cells across ranks and cycles based on their memory load.

        This function uses a greedy algorithm to distribute cells across ranks and cycles such that
        the total memory load is balanced. Cells with higher memory load are distributed first.

        Args:
            num_ranks (int): The number of ranks.
            cycles (int): The number of cycles to distribute cells over.
            batch_size (int): The number of cells to assign to each bucket at a time.

        Returns:
            bucket_allocation (dict): A dictionary where keys are tuples (pop, rank_id, cycle_id)
                                    and values are lists of cell IDs assigned to each bucket.
            bucket_memory (dict): A dictionary where keys are tuples (pop, rank_id, cycle_id)
                                and values are the total memory load on each bucket.
            metype_memory_usage (dict): A dictionary where keys are METype IDs
                                        and values are the memory load of each METype.
        """
        self.validate_inputs_distribute(num_ranks, batch_size)
        bucket_allocation = defaultdict(DryRunStats.defaultdict_vector)
        bucket_memory = defaultdict(DryRunStats.defaultdict_float)

        # Prepare the memory usage for each METype
        metype_memory_usage = {}
        for metype, metype_mem in self.metype_memory.items():
            syns_mem = SynapseMemoryUsage.get_memory_usage(self.metype_cell_syn_average[metype])
            metype_memory_usage[metype] = metype_mem + syns_mem

        def assign_cells_to_bucket(rank_allocation, rank_memory, batch, batch_memory):
            total_memory, (rank_id, cycle_id) = heapq.heappop(buckets)
            logging.debug("Assigning batch to bucket (%d, %d)", rank_id, cycle_id)
            rank_allocation[rank_id, cycle_id].extend(batch)
            total_memory += batch_memory
            rank_memory[rank_id, cycle_id] = total_memory
            heapq.heappush(buckets, (total_memory, (rank_id, cycle_id)))

        # Loop over ALL the gids which would be instantiated, per metype
        for pop, metype_gids in self.pop_metype_gids.items():
            logging.info("Distributing cells of population %s", pop)
            rank_allocation = defaultdict(Vector)
            rank_memory = {}
            batch = []
            batch_memory = 0

            # (total_memory, (rank_id, cycle_id))
            buckets = [(0, (i, j)) for i in range(num_ranks) for j in range(cycles)]
            heapq.heapify(buckets)

            for metype, gids in metype_gids.items():
                total_mem_per_cell = metype_memory_usage[metype]

                for cell_id in gids:
                    batch.append(cell_id)
                    batch_memory += total_mem_per_cell
                    if len(batch) == batch_size[pop]:
                        assign_cells_to_bucket(rank_allocation, rank_memory, batch, batch_memory)
                        batch = []
                        batch_memory = 0

            if batch:
                assign_cells_to_bucket(rank_allocation, rank_memory, batch, batch_memory)

            bucket_allocation[pop] = rank_allocation
            bucket_memory[pop] = rank_memory
        return bucket_allocation, bucket_memory, metype_memory_usage

    def validate_inputs_distribute(self, num_ranks, batch_size):
        assert isinstance(num_ranks, int), "num_ranks must be an integer"
        assert num_ranks > 0, "num_ranks must be a positive integer"
        assert isinstance(batch_size, dict), "batch_size must be a dict"
        assert all(size > 0 for size in batch_size.values()), "batch_size must be positive"

        all_metypes = set()
        for values in self.pop_metype_gids.values():
            all_metypes.update(values.keys())

        # Assert our simulated metypes are a subset of the (pre-) computed ones
        cell_memory_metypes = set(self.metype_memory.keys())
        assert all_metypes <= cell_memory_metypes, all_metypes - cell_memory_metypes

        # NOTE: We shall not assume that we have synapses counts for all metypes
        #       According to override blocks, some cells may be disconnected.
        #       Also Counter() will discard 0-count entries
        # syn_count_metypes = set(self.metype_cell_syn_average)
        # assert all_metypes <= syn_count_metypes, all_metypes - syn_count_metypes

    @staticmethod
    def check_all_buckets_have_gids(bucket_allocation, population, num_ranks, cycles):
        """Checks if all possible buckets determined by num_ranks and cycles have at least one GID
        assigned.

        Args:
            bucket_allocation (dict): The allocation dictionary containing the assignment of GIDs
                                      to ranks and cycles.
            population (str): The population to check.
            num_ranks (int): The number of ranks.
            cycles (int): The number of cycles.
        """
        rank_allocation = bucket_allocation.get(population, {})
        for rank_id in range(num_ranks):
            for cycle_id in range(cycles):
                if not rank_allocation.get((rank_id, cycle_id)):
                    logging.warning(
                        "Population %s is not allocated across the full size of ranks "
                        "and cycles. Consider reducing the number of ranks or cycles.",
                        population,
                    )

    @run_only_rank0
    def distribute_cells_with_validation(self, num_ranks, cycles=None) -> tuple[dict, dict, dict]:
        """Wrapper function to distribute cells across the specified number of ranks and cycles,
        ensuring that each bucket (combination of rank and cycle) has at least one GID assigned.
        The function attempts to find a valid distribution with the initially calculated batch
        sizes, raising an error if no distribution is found.

        Args:
            num_ranks (int): The number of ranks.
            cycles (int): The number of cycles to distribute cells over.

        Returns:
            Tuple[dict, dict, dict]: Returns the same as distribute_cells once a valid distribution
                                    is found.
        """
        if not cycles:
            cycles = 1
        logging.info("Distributing cells across %d ranks and %d cycles", num_ranks, cycles)

        def _calculate_total_elements_per_population(self):
            return {
                population: sum(len(gids) for gids in metypes.values())
                for population, metypes in self.pop_metype_gids.items()
            }

        total_cells_per_population = _calculate_total_elements_per_population(self)
        average_cells_per_population = {
            population: total / (num_ranks * cycles)
            for population, total in total_cells_per_population.items()
        }

        batch_size = {
            population: max(1, math.ceil(average / MAX_ALLOCATION_STEPS))
            for population, average in average_cells_per_population.items()
        }

        bucket_allocation, bucket_memory, metype_memory_usage = self.distribute_cells(
            num_ranks, cycles, batch_size=batch_size
        )

        for population in self.pop_metype_gids:
            self.check_all_buckets_have_gids(bucket_allocation, population, num_ranks, cycles)

        print_allocation_stats(bucket_memory)
        export_allocation_stats(bucket_allocation, self._ALLOCATION_FILENAME, num_ranks, cycles)
        return bucket_allocation, bucket_memory, metype_memory_usage
